(* David Monniaux, CNRS / Laboratoire d'Informatique de                *)
(*                 l'Ecole Normale Superieure                          *)

let debug_msg s = ()

type key = int

let nbits = 29
let test_bit x = (x land (1 lsl (pred nbits))) <> 0
let next_bit x = (x lsl 1) land (pred (1 lsl nbits))

type 'a t =
  Leaf of 'a 
| Node of ('a t) * ('a t)
| Empty;;

let empty = Empty;;

let create n1 n2 =
  match n1, n2 with
(* (Leaf l1), (Leaf l2) when l1==l2 -> Leaf l1 *)
| Empty, Empty -> Empty
| _ -> Node(n1, n2);;

let rec add_rec depth x data map =
  match depth, map with
    0, ((Leaf _) | Empty) ->
      assert(x = 0);
      Leaf data
  | 0, (Node _) -> assert false
  | _ when depth < 0 -> assert false
  | _, Node(b0, b1) ->
      let nx = next_bit x and ndepth = pred depth in
      if test_bit x
      then create b0 (add_rec ndepth nx data b1)
      else create (add_rec ndepth nx data b0) b1
  | _, (Empty | Leaf _) ->
      let nx = next_bit x and ndepth = pred depth in
      if test_bit x
      then create map (add_rec ndepth nx data map)
      else create (add_rec ndepth nx data map) map;;

let add x data map =
  debug_msg "BddMap.add\n";
  if x >= (1 lsl nbits)
  then raise (Invalid_argument "bddMap.add")
  else add_rec nbits x data map;;

let rec find x = function
    Empty -> raise Not_found
  | Leaf data -> data
  | Node(b0, b1) ->
      let nx = next_bit x in
      if test_bit x
      then find nx b1
      else find nx b0;;

let rec mem x = function
    Empty -> false
  | Leaf _ -> true
  | Node(b0, b1) ->
      let nx = next_bit x in
      if test_bit x
      then mem nx b1
      else mem nx b0;;

let rec remove_rec depth x map =
  match depth, map with
    0, ((Leaf _) | Empty) ->
      assert(x = 0);
      Empty
  | 0, (Node _) -> assert false
  | _, Empty -> Empty
  | _, Node(b0, b1) ->
      let nx = next_bit x and ndepth = pred depth in
      if test_bit x
      then create b0 (remove_rec ndepth nx b1)
      else create (remove_rec ndepth nx b0) b1
  | _, (Leaf _) ->
      let nx = next_bit x and ndepth = pred depth in
      if test_bit x
      then create map (remove_rec ndepth nx map)
      else create (remove_rec ndepth nx map) map

let remove x map =
  debug_msg "BddMap.remove\n";
  remove_rec nbits x map;;

(* let rec iter f = function *)
(*     Empty -> () *)
(*   | Leaf data -> f data *)
(*   | Node(b0, b1)-> iter f b0; iter f b1;; *)

let rec iteri_rec path mask f = function
    Empty -> ()
  | Leaf data -> f path data
  | Node(b0, b1) ->
      let nmask = mask lsr 1 in
      iteri_rec path nmask f b0;
      iteri_rec (path lor nmask) nmask f b1

let iter f = iteri_rec 0 (1 lsl nbits) f;;

let rec map f m =
  debug_msg "BddMap.map\n";
  match m with
    Empty -> Empty
  | Leaf data -> Leaf (f data)
  | Node(b0, b1) -> create (map f b0) (map f b1);;

let rec mapi_rec path mask f = function
    Empty -> Empty
  | Leaf data -> Leaf(f path data)
  | Node(b0, b1) ->
      let nmask = mask lsr 1 in
      create (mapi_rec path nmask f b0)
	(mapi_rec (path lor nmask) nmask f b1);;

let mapi f m =
  debug_msg "BddMap.mapi\n";
  mapi_rec 0 (1 lsl nbits) f m;;

let rec fold_rec path mask f map accu =
  match map with
    Empty -> accu
  | Leaf data -> f path data accu
  | Node(b0, b1) ->
      let nmask = mask lsr 1 in
      fold_rec path nmask f b0 (fold_rec (path lor nmask) nmask f b1 accu);;

let fold f map accu = fold_rec 0 (1 lsl nbits) f map accu;;

let rec map2iz_rec path mask f m1 m2 =
  if m1 == m2
  then m1
  else
    match m1, m2 with
      Empty, Empty -> Empty
    | (Leaf data1), (Leaf data2) -> Leaf(f path data1 data2)
    | Node(b10, b11), Node(b20, b21) ->
      let nmask = mask lsr 1 in
      create (map2iz_rec path nmask f b10 b20)
	(map2iz_rec (path lor nmask) nmask f b11 b21)
    | _ ->  raise (Invalid_argument "bddMap.map2iz");;

let map2iz f map1 map2 =
  debug_msg "BddMap.map2iz\n";
  map2iz_rec 0 (1 lsl nbits) f map1 map2;;	

let rec map2i_rec path mask f1 f2 f m1 m2 =
  if m1 == m2
  then m1
  else
    match m1, m2 with
      Empty, Empty -> Empty
    | (Leaf data1), (Leaf data2) -> Leaf(f path data1 data2)
    | Node(b10, b11), Node(b20, b21) ->
      let nmask = mask lsr 1 in
      create (map2i_rec path nmask f1 f2 f b10 b20)
	(map2i_rec (path lor nmask) nmask f1 f2 f b11 b21)
    | (Node _), (Leaf _)
    | (Leaf _), (Node _) -> assert false
    | (Leaf n1), Empty -> Leaf(f1 path n1)
    | Empty, (Leaf n2) -> Leaf(f2 path n2)
    | (Node _), Empty -> mapi_rec path mask f1 m1
    | Empty, (Node _) -> mapi_rec path mask f2 m2;;

let map2i f1 f2 f map1 map2 =
  debug_msg "BddMap.map2i\n";
  map2i_rec 0 (1 lsl nbits) f1 f2 f map1 map2;;

let rec map2_rec path mask f1 f2 f m1 m2 =
   match m1, m2 with
   Empty, Empty -> Empty
   | (Leaf data1), (Leaf data2) -> Leaf(f path data1 data2)
   | Node(b10, b11), Node(b20, b21) ->
   let nmask = mask lsr 1 in
   create (map2_rec path nmask f1 f2 f b10 b20)
   (map2_rec (path lor nmask) nmask f1 f2 f b11 b21)
   | (Node _), (Leaf _)
   | (Leaf _), (Node _) -> assert false
   | (Leaf n1), Empty -> Leaf(f1 path n1)
   | Empty, (Leaf n2) -> Leaf(f2 path n2)
   | (Node _), Empty -> mapi_rec path mask f1 m1
   | Empty, (Node _) -> mapi_rec path mask f2 m2;;

let map2 f1 f2 f map1 map2 =
  debug_msg "BddMap.map2\n";
  map2_rec 0 (1 lsl nbits) f1 f2 f map1 map2;;

let rec iter2iz_rec path mask f m1 m2 =
  if m1 == m2
  then ()
  else
    match m1, m2 with
      Empty, Empty -> ()
    | (Leaf data1), (Leaf data2) -> f path data1 data2
    | Node(b10, b11), Node(b20, b21) ->
      let nmask = mask lsr 1 in
        (iter2iz_rec path nmask f b10 b20);
	(iter2iz_rec (path lor nmask) nmask f b11 b21)
    | _ ->  raise (Invalid_argument "bddMap.iter2iz");;

let iter2iz f map1 map2 = iter2iz_rec 0 (1 lsl nbits) f map1 map2;;	

let rec iter2i_rec path mask f1 f2 f m1 m2 =
  if m1 == m2
  then ()
  else
    match m1, m2 with
      Empty, Empty -> ()
    | (Leaf data1), (Leaf data2) -> f path data1 data2
    | Node(b10, b11), Node(b20, b21) ->
      let nmask = mask lsr 1 in
        (iter2i_rec path nmask f1 f2 f b10 b20);
	(iter2i_rec (path lor nmask) nmask f1 f2 f b11 b21)
    | (Node _), (Leaf _)
    | (Leaf _), (Node _) -> assert false
    | (Leaf n1), Empty -> f1 path n1
    | Empty, (Leaf n2) -> f2 path n2
    | (Node _), Empty -> iteri_rec path mask f1 m1
    | Empty, (Node _) -> iteri_rec path mask f2 m2;;
  
let iter2i f1 f2 f map1 map2 = iter2i_rec 0 (1 lsl nbits) f1 f2 f map1 map2;;

let rec forall2iz_rec path mask f m1 m2 =
  if m1 == m2
  then true
  else
    match m1, m2 with
      Empty, Empty -> true
    | (Leaf data1), (Leaf data2) -> f path data1 data2
    | Node(b10, b11), Node(b20, b21) ->
      let nmask = mask lsr 1 in
        (forall2iz_rec path nmask f b10 b20) &&
	(forall2iz_rec (path lor nmask) nmask f b11 b21)
    | _ ->  raise (Invalid_argument "bddMap.forall2iz");;

let forall2iz f map1 map2 = forall2iz_rec 0 (1 lsl nbits) f map1 map2;;	

(* let foo = List.fold_right (fun (x, y) map -> add x y map) *)
(*   [11, 12; 9, 10; 1, 2; 3, 4; 5, 6; 7, 8] empty;; *)
(* let foo2 = List.fold_right (fun (x, y) map -> add x y map) *)
(*   [3, 6; 13, 14] empty;; *)

(* iteri (fun x y -> Printf.printf "%d %d\n" x y) foo;; *)
