(**
Clustering system
@author David Monniaux
*)

(*
Copyright (c) 2004 David Monniaux,
CNRS / Laboratoire d'informatique de l'Ecole normale superieure
*)

(** Structure describing a computational host. *)
type computation_host = {
  host_address : Unix.sockaddr option;
  host_speed : float;
}

(** Syntax error in cluster description file *)
exception Syntax_error of string * int

(** Load a list of computational hosts from a cluster description file.
Each line in the file is of the format:
[name speed]
or
[name speed host]
or
[name speed host:port]
where [name] is the name of the machine, [speed] is a floating-point number expressing its relative speed, [host] is a host name or a IP address, and [port] is a TCP port number.

[load_computation_hosts_from_file port filename] loads the file [filename] assuming that the default port is [port].*)
val load_computation_hosts_from_file :
  int -> string -> computation_host list

(** Call a remote function [f] by RPC over a cluster. *)
val parallel_call : 'c ->
  ('c * ('a list) -> 'b) ->
  ('c * ('a list), 'b) RPC.magic_token ->
  computation_host list -> 'a list -> 'b list

val ping : ('a,'b) RPC.magic_token -> computation_host list -> unit
