open Parsing_helper

(* Modules for Ordered tries (maps and sets) structures *)

module PLocusOrd  = struct type t = pseudo_locus let compare = compare        end
module StringOrd  = struct type t = string       let compare = String.compare end
module IntegerOrd = struct type t = int          let compare = ( - )          end
module LocusOrd   = struct type t = locus        let compare = compare        end
module ExtentOrd  = struct type t = extent       let compare = compare        end
module PLocusMap  = Map2.Make(PLocusOrd)
module PLocusSet  = Set.Make(PLocusOrd)
module StringMap  = Map2.Make(StringOrd)
module StringSet  = Set.Make(StringOrd)
module IntegerMap = Map2.Make(IntegerOrd)
module LocusMap   = Map2.Make(LocusOrd)
module LocusSet   = Set.Make(LocusOrd)
module ExtentMap  = Map2.Make(ExtentOrd)
module IntegerSet =
  struct
    include Ptset (*Set.Make(struct type t=int let compare = ( - ) end)*) (* Ptset *)
    let of_list : elt list -> t = function l ->
      List.fold_right add l empty
  end
module KeyMap = Maps.Product (IntegerMap) (IntegerMap)
module KeySet = Sets.DisjointUnion (IntegerSet) (IntegerSet)

module BigIntegerSet = Set.Make(struct type t=int64 let compare=compare end);;

(* Some printers for sets *)
(*  xr: we may build them into the module definitions *)
let pp_intset (chan: out_channel) (s: IntegerSet.t): unit =
  let b_first = ref true in
  Printf.fprintf chan "{";
  IntegerSet.iter
    (fun i ->
      if !b_first then
	begin
	  Printf.fprintf chan " %d" i;
	  b_first := false;
	end
      else Printf.fprintf chan "; %d" i;
    ) s;
  Printf.fprintf chan " }"
let pp_stringset (chan: out_channel) (s: StringSet.t): unit =
  let b_first = ref true in
  Printf.fprintf chan "{";
  StringSet.iter
    (fun str ->
      if !b_first then
	begin
	  Printf.fprintf chan " %s" str;
	  b_first := false;
	end
      else Printf.fprintf chan "; %s" str;
    ) s;
  Printf.fprintf chan " }"

(* Some map functions *)

let rec map_rec f others = function
    [] -> others
  | h::t -> map_rec f ((f h)::others) t

let rec revconcat others = function
    [] -> others
  | h::t -> revconcat (h::others) t

let rec map_and_flatten f = function
    [] -> []
  | h::t -> revconcat (f h) (map_and_flatten f t)

let array_map2 f a1 a2 =
  assert((Array.length a1) = (Array.length a2));
  Array.mapi (fun i x -> f x a2.(i)) a1

let rec quad_map f l1 l2 = match l1 with
  [] -> []
| x::l1' -> (List.map (f x) l2)@(quad_map f l1' l2)

let rec even_odd = function
    [] -> [], []
  | [_] as single -> single, []
  | h1::h2::t -> let t1, t2 = even_odd t in
    (h1 :: t1), (h2 :: t2);;

let pp_option pp chan = function
  | None -> Printf.fprintf chan "None"
  | Some y -> pp chan y
