let change_tresh = 10;;

module IntegerMap = Map.Make
    (struct
      type t=int
      let compare (x : t) (y : t) = Pervasives.compare x y
    end)

type key=int
type size=int

type 'a t={
    array : 'a array;
    deltas : 'a IntegerMap.t;
    changes : int;
    length : int }

let init n f =
  { array = Array.init n f;
    deltas = IntegerMap.empty;
    changes = 0;
    length = n }

let get a k =
  (if k < 0 || k > a.length
  then invalid_arg "FArray.get");
  try IntegerMap.find k a.deltas
  with Not_found -> Array.get a.array k

let find k a = get a k

let iter f a =
  for i=0 to pred a.length
  do
    f (get a i)
  done

let iteri f a =
  for i=0 to pred a.length
  do
    f i (get a i)
  done

let squash_array a = Array.init a.length (get a)

let simplify a =
  if a.changes >= change_tresh
  then { array = squash_array a;
	 deltas = IntegerMap.empty;
	 changes = 0;
	 length = a.length }
  else a

let empty =
  { array = [| |];
    deltas = IntegerMap.empty;
    changes =0;
    length = 0 }

let create n x =
  { array = Array.create n x;
    deltas = IntegerMap.empty;
    changes = 0;
    length = n }

let make = create

let length a = a.length

let add k x a =
  if k <> a.length
  then invalid_arg "FArray.add"
  else simplify
    { array = a.array;
      deltas = IntegerMap.add k x a.deltas;
      changes = succ a.changes;
      length = succ a.length }

let remove k a =
  if k <> (pred a.length)
  then invalid_arg "FArray.remove"
  else
    { array = a.array;
      deltas = a.deltas;
      changes = succ a.changes;
      length = k }

let set a k x =
  if k < 0 || k >= a.length
  then invalid_arg "FArray.set"
else simplify
    { array = a.array;
      deltas = IntegerMap.add k x a.deltas;
      changes = succ a.changes;
      length = a.length }

let map f a = init a.length (fun i -> f (get a i))
let mapi f a = init a.length (fun i -> f i (get a i))

let map2 f a b =
  (if a.length <> b.length then invalid_arg "FArray.map2");
  init a.length (fun i -> f (get a i) (get b i))
let mapi2 f a b =
  (if a.length <> b.length then invalid_arg "FArray.map2");
  init a.length (fun i -> f i (get a i) (get b i))
  
let pretty_print item_printer formatter a =
  Format.fprintf formatter "@[<><| l=%d c=%d" a.length a.changes;
  iter
    (fun x ->
      Format.pp_print_string formatter "; ";
      item_printer formatter x) a;
  Format.fprintf formatter " |>@]"

(* let pp = (pretty_print Format.pp_print_int);; *)

(* #install_printer pp;; *)

(* let a1 = init 10 (fun i -> i);; *)
(* let a2 = add 10 110 a1;; *)
(* let a3 = set a2 10 210;; *)
(* let a4 = set a3 0 200;; *)
(* let a5 = set a4 1 201;; *)
