/* Monomorphic comparison functions.
   I hope these are faster than Pervasives.compare !

   Copyright (c) Antoine Mine' 2003
*/


#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>
#include <caml/custom.h>
#include <caml/fail.h>

#include <assert.h>
#include <math.h>

/* inline for faster comparison... */
#ifndef ARCH_ALIGN_INT64
#define my_Int64_val(v) (*((int64 *) Data_custom_val(v)))
#else
static inline int64 my_Int64_val(value v)
{
  union { int32 i[2]; int64 j; } buffer;
  buffer.i[0] = ((int32 *) Data_custom_val(v))[0];
  buffer.i[1] = ((int32 *) Data_custom_val(v))[1];
  return buffer.j;
}
#endif

#ifndef ARCH_ALIGN_DOUBLE
#define my_Double_val(v) (* (double *)(v))
#else
static inline double my_Double_val(value val)
{
  union { value v[2]; double d; } buffer;
  buffer.v[0] = Field(val, 0);
  buffer.v[1] = Field(val, 1);
  return buffer.d;
}
#endif

CAMLprim value ens_int_compare(value v1, value v2)
{
  CAMLparam2(v1,v2);
  CAMLreturn(v1 == v2 ? Val_int(0) : v1 < v2 ? Val_int(-1) : Val_int(1));
}

extern struct custom_operations caml_int64_ops;

CAMLprim value ens_int64_compare(value v1, value v2)
{
  CAMLparam2(v1,v2);
  int64 i1,i2; 
  /*assert(Custom_ops_val(v1)==&caml_int64_ops);*/
  /*assert(Custom_ops_val(v2)==&caml_int64_ops);*/
  i1 = my_Int64_val(v1);
  i2 = my_Int64_val(v2);
  CAMLreturn(i1 == i2 ? Val_int(0) : i1 < i2 ? Val_int(-1) : Val_int(1));
}

CAMLprim value ens_float_compare(value v1, value v2)
{
  CAMLparam2(v1,v2);
  double i1,i2;
  i1 = my_Double_val(v1);
  i2 = my_Double_val(v2);
  CAMLreturn(i1 == i2 ? Val_int(0) : i1 < i2 ? Val_int(-1) : Val_int(1));
}


CAMLprim value ens_key_compare(value v1, value v2)
{
  CAMLparam2(v1,v2);
  long t1 = Long_val(Field(v1,0));
  long t2 = Long_val(Field(v2,0));
  long r;
  if (t1==t2) r = Long_val(Field(v1,1))-Long_val(Field(v2,1));
  else r = t1-t2;
  CAMLreturn(r<0 ? Val_int(-1) : r>0 ? Val_int(1) : Val_int(0));
}

CAMLprim value ens_string_compare(value v1, value v2)
{
  CAMLparam2(v1,v2);
  mlsize_t len1, len2, len;
  unsigned char * p1, * p2;
  long r;
  len1 = string_length(v1);
  len2 = string_length(v2);
  r = len1-len2;
  for (len = (len1 <= len2 ? len1 : len2),
	 p1 = (unsigned char *) String_val(v1),
	 p2 = (unsigned char *) String_val(v2);
       len > 0;
       len--, p1++, p2++)
    if (*p1 != *p2) { r = (long)*p1 - (long)*p2; break; }
  CAMLreturn(r<0? Val_int(-1) : r>0 ? Val_int(1) : Val_int(0));
}

