(*
Copyright (c) 2004 David Monniaux,
CNRS / Laboratoire d'informatique de l'Ecole normale superieure
*)

let rec prepend_to_list n item tail =
  assert(n >= 0);
  match n with
    0 -> tail
  | _ -> item::(prepend_to_list (pred n) item tail);;

let rec length_is_at_least n l =
  if (n <= 0)
  then true
  else
    match l with
      _::t -> length_is_at_least (pred n) t
    | [] -> false;;

let rec extract_head_items n l =
  match n, l with
    n, _ when n < 0 -> raise (Invalid_argument "extract_head_items")
  | 0, t -> [], t
  | _, [] -> raise (Invalid_argument "extract_head_items")
  | n, h::t -> let h', t' = extract_head_items (pred n) t in (h::h'), t';;

let rec divide_list_in_n n l =
  assert(n > 0);
  if length_is_at_least n l
  then
    let heads, tail = extract_head_items n l in
    let tails = divide_list_in_n n tail in
    List.map2 (fun x y -> x::y) heads tails
  else
    prepend_to_list (n - (List.length l)) []
      (List.map (fun x -> [x]) l);;

let rec split_list_by_numbers numbers items =
  match numbers, items with
    [], [] -> []
  | [], _ -> raise (Invalid_argument "split_list_by_numbers")
  | hn::tn, _ ->
      let nh, nt = extract_head_items hn items in
      nh::(split_list_by_numbers tn nt);;

let shuffle_array a =
  let n = Array.length a in
  for i=0 to n-2
  do
    let j = i + (Random.int (n - i)) in
    let tmp = a.(j) in
    a.(j) <- a.(i);
    a.(i) <- tmp
  done;;

let shuffle_list l =
  let a = Array.of_list l in
  shuffle_array a;
  Array.to_list a;;

let map f l =
  List.rev (List.rev_map f l)

let fold_right f l a =
  List.fold_left (fun x y -> f y x) a (List.rev l)
