(**
List utility functions
@author David Monniaux
*)

(*
Copyright (c) 2004 David Monniaux,
CNRS / Laboratoire d'informatique de l'Ecole normale superieure
*)

(** [prepend_to_list n item l] prepends [n] times [item] to the list [l] *)
val prepend_to_list : int -> 'a -> 'a list -> 'a list

(** [length_is_at_least n l] evaluates to [true] if and only if the length of [l] is greater or equal to [n]. *)
val length_is_at_least : int -> 'a list -> bool

(** [extract_head_items n l] yields a pair ([l1], [l2]) where [l1] is the first [n] items of [l] and [l2] is the remainder.
@raise Invalid_argument if [n] is greater than the length of [l]. *)
val extract_head_items : int -> 'a list -> 'a list * 'a list

(** [divide_list_in_n n l] yields a list of [n] lists of approximately [(List.length l)]/[n] length. *)
val divide_list_in_n : int -> 'a list -> 'a list list

(** [split_list_by_numbers [n1; n2; ...] l] splits [l] into lists of length n1, n2, ... *)
val split_list_by_numbers : int list -> 'a list -> 'a list list

(** [shuffle_list l] returns [l] randomly shuffled (random permutation,
  uniformly distributed) *)
val shuffle_list : 'a list -> 'a list

(** recursif [AM]. *)
val map: ('a -> 'b) -> 'a list -> 'b list
val fold_right: ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
