#include <ppl.hh>
#include <iostream>
#include "ml_interface.h"
#include <gmp_caml.h>

#define MODULE "PPL"

#define CAMLCALL(id) extern "C" value ml_ppl_##id

namespace PPL = Parma_Polyhedra_Library;
using namespace PPL::IO_Operators;
using namespace PPL;

static inline value value_of_mpz_class(const mpz_class& input) {
  mpz_t local;
  mpz_init_set(local, input.get_mpz_t());
  mpz_ptr localp = local;
  return camlidl_mpz_ptr_c2ml(&localp);
}

#define CPP_TYPE(caml_type, Cpp_Type) \
static inline Cpp_Type& caml_type##_of_value(value r) {	\
  return *((Cpp_Type*) Data_custom_val(r));			\
}								\
								\
extern "C" void ml_ppl_##caml_type##_finalize(value r) {	\
  caml_type##_of_value(r).~Cpp_Type();				\
}								\
								\
static struct custom_operations ml_ppl_##caml_type##_custom = {	\
  "PPL." #caml_type,						\
  &ml_ppl_##caml_type##_finalize,                               \
  NULL, NULL, NULL, NULL				        \
};

#define CPP_TYPE_DEF_CREATE(caml_type, Cpp_Type)			\
static value caml_type##_create(Cpp_Type expr) {		\
  value r = caml_alloc_custom(&ml_ppl_##caml_type##_custom,	\
		   sizeof(Cpp_Type),				\
		   0,						\
		   1);						\
  new(&caml_type##_of_value(r)) Cpp_Type(expr);			\
  return r;							\
}

#define CPP_TYPE_AND_CREATE(caml_type, Cpp_Type)\
  CPP_TYPE(caml_type, Cpp_Type) \
  CPP_TYPE_DEF_CREATE(caml_type, Cpp_Type)

CPP_TYPE_AND_CREATE(linear_expr, Linear_Expression)
CPP_TYPE(nnc_poly, NNC_Polyhedron)
CPP_TYPE_AND_CREATE(constraint, Constraint)
CPP_TYPE(lp_problem, LP_Problem)
CPP_TYPE_AND_CREATE(generator, Generator)

CAMLCALL(linear_expr_from_variable)(value variable) {
  CAMLparam1(variable);
  CAMLreturn(linear_expr_create(Variable(Int_val(variable))));
}

CAMLCALL(linear_expr_from_scaled_variable)(value _v_coeff, value variable) {
  CAMLparam2(_v_coeff, variable);
  mpz_ptr coeff;
  camlidl_mpz_ptr_ml2c(_v_coeff, &coeff);
  CAMLreturn(linear_expr_create(mpz_class(coeff) * Variable(Int_val(variable))));
}

CAMLCALL(linear_expr_println)(value expr) {
  CAMLparam1(expr);
  std::cout << linear_expr_of_value(expr) << "\n";
  std::cout.flush();
  CAMLreturn(Val_unit);
}

CAMLCALL(linear_expr_scale)(value _v_coeff, value expr) {
  CAMLparam2(_v_coeff, expr);
  mpz_ptr coeff;
  camlidl_mpz_ptr_ml2c(_v_coeff, &coeff);  
  CAMLreturn(linear_expr_create(mpz_class(coeff) * linear_expr_of_value(expr)));}

CAMLCALL(linear_expr_from_z)(value _v_coeff) {
  CAMLparam1(_v_coeff);
  CAMLlocal1(_vres);

  mpz_ptr coeff;
  camlidl_mpz_ptr_ml2c(_v_coeff, &coeff);
  CAMLreturn(linear_expr_create(Linear_Expression(mpz_class(coeff))));
}

CAMLCALL(linear_expr_add)(value e1, value e2) {
  CAMLparam2(e1, e2);
  CAMLreturn(linear_expr_create(linear_expr_of_value(e1) + linear_expr_of_value(e2)));
}

CAMLCALL(constraint_make)(value e1, value kind, value e2) {
  CAMLparam3(e1, kind, e2);
  CAMLlocal1(r);
  switch (kind) {
  case 1:
    r = constraint_create(linear_expr_of_value(e1) < linear_expr_of_value(e2));
    break;
  case 3:
    r = constraint_create(linear_expr_of_value(e1) <= linear_expr_of_value(e2));
    break;
  case 5:
    r = constraint_create(linear_expr_of_value(e1) > linear_expr_of_value(e2));
    break;
  case 7:
    r = constraint_create(linear_expr_of_value(e1) >= linear_expr_of_value(e2));
    break;
  case 9:
    r = constraint_create(linear_expr_of_value(e1) == linear_expr_of_value(e2));
    break;
  default:
    caml_failwith("PPL.constraint_make bad kind value");
  }
  CAMLreturn(r);
}

CAMLCALL(constraint_println)(value constraint) {
  CAMLparam1(constraint);
  std::cout << constraint_of_value(constraint) << "\n";
  std::cout.flush();
  CAMLreturn(Val_unit);
}

CAMLCALL(nnc_poly_create)(value dimension, value degenerate) {
  CAMLparam2(dimension, degenerate);
  CAMLlocal1(r);
  r = caml_alloc_custom(&ml_ppl_nnc_poly_custom,
		   sizeof(NNC_Polyhedron),
		   0,
		   1);
  new(&nnc_poly_of_value(r)) NNC_Polyhedron(Int_val(dimension),
					    degenerate == 1 ? UNIVERSE : EMPTY);
  CAMLreturn(r);
}

CAMLCALL(nnc_poly_add_constraint)(value p, value c) {
  CAMLparam2(p, c);
  nnc_poly_of_value(p).add_constraint(constraint_of_value(c));
  CAMLreturn(Val_unit);
}

CAMLCALL(nnc_poly_add_constraint_and_minimize)(value p, value c) {
  CAMLparam2(p, c);
  nnc_poly_of_value(p).add_constraint_and_minimize(constraint_of_value(c));
  CAMLreturn(Val_unit);
}

CAMLCALL(nnc_poly_println)(value expr) {
  CAMLparam1(expr);
  std::cout << nnc_poly_of_value(expr) << "\n";
  std::cout.flush();
  CAMLreturn(Val_unit);
}

CAMLCALL(nnc_poly_remove_higher_space_dimensions)(value p, value d) {
  CAMLparam2(p, d);
  nnc_poly_of_value(p).remove_higher_space_dimensions(Int_val(d));
  CAMLreturn(Val_unit);
}

CAMLCALL(nnc_poly_add_constraints)(value p, value clist) {
  CAMLparam2(p, clist);
  CAMLlocal1(c);
  Constraint_System cs;
  c = clist;
  while (c != 1) {
    cs.insert(constraint_of_value(Field(c, 0)));
    c = Field(c, 1);
  }
  nnc_poly_of_value(p).add_constraints(cs);
  CAMLreturn(Val_unit);
}

CAMLCALL(nnc_poly_add_constraints_and_minimize)(value p, value clist) {
  CAMLparam2(p, clist);
  CAMLlocal1(c);
  Constraint_System cs;
  c = clist;
  while (c != 1) {
    cs.insert(constraint_of_value(Field(c, 0)));
    c = Field(c, 1);
  }
  nnc_poly_of_value(p).add_constraints_and_minimize(cs);
  CAMLreturn(Val_unit);
}

CAMLCALL(nnc_poly_constraints)(value p) {
  CAMLparam1(p);
  CAMLlocal3(r, r_old, ct);
  r = 1;
  const Constraint_System& constraints = nnc_poly_of_value(p).constraints();
  for(Constraint_System::const_iterator constraint = constraints.begin(),
        constraint_end = constraints.end();
      constraint != constraint_end;
      ++constraint) {
    ct = constraint_create(*constraint);
    r_old = r;
    r = caml_alloc(2, 0);                   /* Allocate a cons cell */
    Store_field(r, 0, ct);
    Store_field(r, 1, r_old);  
  }
  CAMLreturn(r);
}

CAMLCALL(nnc_poly_minimized_constraints)(value p) {
  CAMLparam1(p);
  CAMLlocal3(r, r_old, ct);
  r = 1;
  const Constraint_System& constraints = nnc_poly_of_value(p).constraints();
  for(Constraint_System::const_iterator constraint = constraints.begin(),
        constraint_end = constraints.end();
      constraint != constraint_end;
      ++constraint) {
    ct = constraint_create(*constraint);
    r_old = r;
    r = caml_alloc(2, 0);                   /* Allocate a cons cell */
    Store_field(r, 0, ct);
    Store_field(r, 1, r_old);  
  }
  CAMLreturn(r);
}

CAMLCALL(nnc_poly_is_empty)(value p) {
  CAMLparam1(p);
  CAMLreturn(Val_bool(nnc_poly_of_value(p).is_empty()));
}

CAMLCALL(nnc_poly_is_universe)(value p) {
  CAMLparam1(p);
  CAMLreturn(Val_bool(nnc_poly_of_value(p).is_universe()));
}

CAMLCALL(constraint_read)(value p) {
  const Constraint& constraint = constraint_of_value(p);
  CAMLparam1(p);
  CAMLlocal3(coeff_list, tmp1, tmp2);

  coeff_list = 1;

  for(int k=constraint.space_dimension()-1; k>=0; k--) {
    Coefficient coeff(constraint.coefficient(Variable(k)));
    if (coeff != 0) {
      tmp1 = value_of_mpz_class(coeff);
      tmp2 = caml_alloc(2, 0); /* Allocate a cons cell */
      Store_field(tmp2, 0, Val_int(k));
      Store_field(tmp2, 1, tmp1);

      tmp1 = caml_alloc(2, 0); /* Allocate a cons cell */
      Store_field(tmp1, 0, tmp2);
      Store_field(tmp1, 1, coeff_list);
      coeff_list = tmp1;
    }
  }

  tmp1 = value_of_mpz_class(mpz_class(-constraint.inhomogeneous_term()));
  tmp2 = caml_alloc(3, 0);
  Store_field(tmp2, 0, coeff_list);
  Store_field(tmp2, 2, tmp1);

  switch (constraint.type()) {
    case Constraint::EQUALITY:
      Store_field(tmp2, 1, 9);
      break;
    case Constraint::NONSTRICT_INEQUALITY:
      Store_field(tmp2, 1, 7);
      break;
    case Constraint::STRICT_INEQUALITY:
      Store_field(tmp2, 1, 5);
      break;
  default:
    caml_failwith("nnc_constraint_read");
  }

  CAMLreturn(tmp2);
}

CAMLCALL(lp_problem_create)(value constraints, value objective, value mode) {
  CAMLparam3(constraints, objective, mode);
  CAMLlocal2(c, r);
  Constraint_System cs;
  c = constraints;
  while (c != 1) {
    cs.insert(constraint_of_value(Field(c, 0)));
    c = Field(c, 1);
  }

  r = caml_alloc_custom(&ml_ppl_lp_problem_custom,
			sizeof(LP_Problem),
			0,
			1);
  new(&lp_problem_of_value(r)) LP_Problem(cs, linear_expr_of_value(objective), mode==1 ? MINIMIZATION : MAXIMIZATION);
  CAMLreturn(r);
}

CAMLCALL(lp_problem_solve)(value problem) {
  CAMLparam1(problem);
  value r;
  switch (lp_problem_of_value(problem).solve()) {
  case UNFEASIBLE_LP_PROBLEM:
    r = 1; break;
  case UNBOUNDED_LP_PROBLEM:
    r = 3; break;
  case OPTIMIZED_LP_PROBLEM:
    r = 5; break;
  default:
    caml_failwith("lp_problem_solve: unexpected return value");
  }
  CAMLreturn(r);
}

CAMLCALL(lp_problem_optimal_value)(value problem) {
  CAMLparam1(problem);
  CAMLlocal1(r);
  Coefficient num, den;
  lp_problem_of_value(problem).optimal_value(num, den);
  
  r = caml_alloc(2, 0);
  
  Store_field(r, 0, value_of_mpz_class(num));
  Store_field(r, 1, value_of_mpz_class(den));

  CAMLreturn(r);
}

CAMLCALL(lp_problem_feasible_point)(value problem) {
  CAMLparam1(problem);
  CAMLreturn(generator_create(lp_problem_of_value(problem).feasible_point()));
}

CAMLCALL(lp_problem_optimizing_point)(value problem) {
  CAMLparam1(problem);
  CAMLreturn(generator_create(lp_problem_of_value(problem).optimizing_point()));
}

#define GENERATOR_IS_XXX(XXX) \
CAMLCALL(generator_is_##XXX)(value v) { \
  CAMLparam1(v); \
  CAMLreturn(Val_bool(generator_of_value(v).is_##XXX())); \
}

GENERATOR_IS_XXX(line)
GENERATOR_IS_XXX(ray)
GENERATOR_IS_XXX(point)
GENERATOR_IS_XXX(closure_point)

CAMLCALL(generator_println)(value v) {
  CAMLparam1(v);
  std::cout << generator_of_value(v) << "\n";
  CAMLreturn(Val_unit);
}

CAMLCALL(generator_read_divisor)(value p) {
  CAMLparam1(p);
  CAMLlocal1(r);
  CAMLreturn(value_of_mpz_class(generator_of_value(p).divisor()));
}

CAMLCALL(generator_read_coefficients)(value p) {
  CAMLparam1(p);
  CAMLlocal3(coeff_list, tmp1, tmp2);
  const Generator& gen = generator_of_value(p);
  coeff_list = 1;
  for(unsigned k=0; k<gen.space_dimension(); k++) {
    Coefficient coeff = gen.coefficient(Variable(k));
    if (coeff != 0) {
      tmp1 = value_of_mpz_class(coeff);
      tmp2 = caml_alloc(2, 0); /* Allocate a cons cell */
      Store_field(tmp2, 0, Val_int(k));
      Store_field(tmp2, 1, tmp1);
      tmp1 = caml_alloc(2, 0); /* Allocate a cons cell */
      Store_field(tmp1, 0, tmp2);
      Store_field(tmp1, 1, coeff_list);
      coeff_list = tmp1;
    }
  }
  CAMLreturn(coeff_list);
}
